@@ -1,6 +1,6 @@
 /* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */
 /*
- *  Copyright (C) 2012-2016  Ruby-GNOME2 Project Team
+ *  Copyright (C) 2012-2019  Ruby-GNOME2 Project Team
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -42,6 +42,14 @@ typedef struct _RBGICallback {
     ffi_closure *closure;
 } RBGICallback;
 
+struct RBGICallbackData_ {
+    RBGICallback *callback;
+    RBGIArgMetadata *metadata;
+    VALUE rb_callback;
+    GObject *owner;
+    VALUE rb_owner;
+};
+
 static VALUE RG_TARGET_NAMESPACE;
 static VALUE rb_cGLibError;
 static const char *callbacks_key = "gi_callbacks";
@@ -127,6 +135,7 @@ allocate_arguments(GICallableInfo *info,
         memset(&argument, 0, sizeof(GIArgument));
 
         metadata = ALLOC(RBGIArgMetadata);
+        metadata->callable_info = info;
         arg_info = &(metadata->arg_info);
         g_callable_info_load_arg(info, i, arg_info);
         metadata->scope_type = g_arg_info_get_scope(arg_info);
@@ -297,27 +306,6 @@ fill_metadata(GPtrArray *args_metadata, GICallableInfo *info)
     fill_metadata_rb_arg_index(args_metadata);
 }
 
-static void
-callback_data_guard_from_gc(RBGICallbackData *callback_data)
-{
-    VALUE rb_callbacks;
-
-    rb_callbacks = rb_iv_get(RG_TARGET_NAMESPACE, callbacks_key);
-    callback_data->rb_gc_guard_key = rb_class_new_instance(0, NULL, rb_cObject);
-    rb_hash_aset(rb_callbacks,
-                 callback_data->rb_gc_guard_key,
-                 callback_data->rb_callback);
-}
-
-static void
-callback_data_unguard_from_gc(RBGICallbackData *callback_data)
-{
-    VALUE rb_callbacks;
-
-    rb_callbacks = rb_iv_get(RG_TARGET_NAMESPACE, callbacks_key);
-    rb_hash_delete(rb_callbacks, callback_data->rb_gc_guard_key);
-}
-
 static void
 rb_gi_callback_free(RBGICallback *callback)
 {
@@ -328,19 +316,54 @@ rb_gi_callback_free(RBGICallback *callback)
     xfree(callback);
 }
 
+static void
+rb_gi_callback_data_weak_notify(gpointer data,
+                                G_GNUC_UNUSED GObject *where_the_object_was)
+{
+    RBGICallbackData *callback_data = data;
+    callback_data->owner = NULL;
+}
+
 void
 rb_gi_callback_data_free(RBGICallbackData *callback_data)
 {
     if (callback_data->callback) {
         rb_gi_callback_free(callback_data->callback);
     }
-    callback_data_unguard_from_gc(callback_data);
+    if (callback_data->owner) {
+        VALUE rb_owner;
+        g_object_weak_unref(callback_data->owner,
+                            rb_gi_callback_data_weak_notify,
+                            callback_data);
+        rb_owner = rbgobj_ruby_object_from_instance2(callback_data->owner,
+                                                     FALSE);
+        if (!NIL_P(rb_owner)) {
+            rbgobj_object_remove_relative(rb_owner, callback_data->rb_callback);
+        }
+    }
+    if (!NIL_P(callback_data->rb_owner)) {
+        rbgobj_remove_relative(callback_data->rb_owner,
+                               (ID)0,
+                               callback_data->rb_callback);
+    }
     xfree(callback_data->metadata);
     xfree(callback_data);
 }
 
+RBGIArgMetadata *
+rb_gi_callback_data_get_metadata(RBGICallbackData *callback_data)
+{
+    return callback_data->metadata;
+}
+
+VALUE
+rb_gi_callback_data_get_rb_callback(RBGICallbackData *callback_data)
+{
+    return callback_data->rb_callback;
+}
+
 static void
-destroy_notify(gpointer data)
+rb_gi_callback_data_destroy_notify(gpointer data)
 {
     RBGICallbackData *callback_data = data;
     rb_gi_callback_data_free(callback_data);
@@ -1023,6 +1046,29 @@ ffi_closure_callback(G_GNUC_UNUSED ffi_cif *cif,
             callback_data = *((RBGICallbackData **)(raw_args[i]));
             break;
         }
+
+        if (!callback_data && args_metadata->len > 0) {
+            RBGIArgMetadata *metadata;
+            GIArgInfo *arg_info;
+            GITypeInfo *type_info;
+            GITypeTag type_tag;
+            gboolean is_pointer;
+            const gchar *arg_name;
+
+            i = args_metadata->len - 1;
+            metadata = g_ptr_array_index(args_metadata, i);
+            arg_info = &(metadata->arg_info);
+            type_info = g_arg_info_get_type(arg_info);
+            type_tag = g_type_info_get_tag(type_info);
+            is_pointer = g_type_info_is_pointer(type_info);
+            g_base_info_unref(type_info);
+            arg_name = g_base_info_get_name(arg_info);
+            if (type_tag == GI_TYPE_TAG_VOID &&
+                is_pointer &&
+                strcmp(arg_name, "data") == 0) {
+                callback_data = *((RBGICallbackData **)(raw_args[i]));
+            }
+        }
     }
 
     {
@@ -1039,7 +1085,7 @@ ffi_closure_callback(G_GNUC_UNUSED ffi_cif *cif,
         rb_results = rb_funcallv(callback_data->rb_callback,
                                  id_call,
                                  rb_args->len,
-                                 (VALUE *)(rb_args->data));
+                                 (VALUE *)(void *)(rb_args->data));
         g_array_free(rb_args, TRUE);
     }
 
@@ -1055,7 +1101,10 @@ ffi_closure_callback(G_GNUC_UNUSED ffi_cif *cif,
 }
 
 static void
-in_callback_argument_from_ruby(RBGIArgMetadata *metadata, GArray *in_args)
+in_callback_argument_from_ruby(RBGIArgMetadata *metadata,
+                               GArray *in_args,
+                               VALUE self,
+                               VALUE rb_arguments)
 {
     gpointer callback_function;
     GIArgInfo *arg_info;
@@ -1108,27 +1157,77 @@ in_callback_argument_from_ruby(RBGIArgMetadata *metadata, GArray *in_args)
 
     if (closure_argument) {
         RBGICallbackData *callback_data;
+        VALUE rb_owner = self;
+        static VALUE mGLibObject = Qnil;
+
+        if (rb_gi_is_debug_mode()) {
+            GICallableInfo *callable_info = metadata->callable_info;
+            g_print("[rb-gi] callback: %s::%s()\n",
+                    g_base_info_get_namespace(callable_info),
+                    g_base_info_get_name(callable_info));
+            rb_p(rb_ary_new_from_args(2, self, rb_arguments));
+        }
 
         callback_data = ALLOC(RBGICallbackData);
         callback_data->callback = callback;
         callback_data->metadata = metadata;
         callback_data->rb_callback = rb_block_proc();
-        callback_data_guard_from_gc(callback_data);
+        if (NIL_P(mGLibObject)) {
+            mGLibObject = rb_const_get(mGLib, rb_intern("Object"));
+        }
+        if (NIL_P(rb_owner)) {
+            /* Module function case. */
+            VALUE rb_first_argument = rb_ary_entry(rb_arguments, 0);
+            if (rb_obj_is_kind_of(rb_first_argument, mGLibObject)) {
+                /* If the first argument of the module function call is
+                   GObject, it's suitable for owner.
+                   For example: pango_cairo_context_set_shape_renderer() */
+                rb_owner = rb_first_argument;
+            }
+        }
+        if (rb_obj_is_kind_of(rb_owner, mGLibObject)) {
+            rbgobj_object_add_relative(rb_owner, callback_data->rb_callback);
+            callback_data->owner = RVAL2GOBJ(rb_owner);
+            g_object_weak_ref(callback_data->owner,
+                              rb_gi_callback_data_weak_notify,
+                              callback_data);
+            callback_data->rb_owner = Qnil;
+        } else {
+            /* Callback is GC-ed only when callback is invalidated. */
+            if (NIL_P(rb_owner)) {
+                /* Module function case. */
+                static VALUE rb_mGI = Qnil;
+                if (NIL_P(rb_mGI)) {
+                    rb_mGI = rb_const_get(rb_cObject,
+                                          rb_intern("GObjectIntrospection"));
+                }
+                rbgobj_add_relative(rb_mGI, callback_data->rb_callback);
+                callback_data->rb_owner = rb_mGI;
+            } else {
+                /* Class method case. */
+                rbgobj_add_relative(rb_owner, callback_data->rb_callback);
+                callback_data->rb_owner = rb_owner;
+            }
+            callback_data->owner = NULL;
+        }
         closure_argument->v_pointer = callback_data;
     }
 
     if (destroy_argument) {
-        destroy_argument->v_pointer = destroy_notify;
+        destroy_argument->v_pointer = rb_gi_callback_data_destroy_notify;
     }
 }
 
 static void
-in_argument_from_ruby(RBGIArgMetadata *metadata, VALUE rb_arguments,
-                      GArray *in_args, GPtrArray *args_metadata,
+in_argument_from_ruby(GICallableInfo *callable_info,
+                      RBGIArgMetadata *metadata,
+                      VALUE rb_arguments,
+                      GArray *in_args,
+                      GPtrArray *args_metadata,
                       VALUE self)
 {
     if (metadata->callback_p && !metadata->destroy_p) {
-        in_callback_argument_from_ruby(metadata, in_args);
+        in_callback_argument_from_ruby(metadata, in_args, self, rb_arguments);
         return;
     }
 
@@ -1172,10 +1271,12 @@ in_argument_from_ruby(RBGIArgMetadata *metadata, VALUE rb_arguments,
             rb_argument = RARRAY_PTR(rb_arguments)[metadata->rb_arg_index];
         }
         argument = &(g_array_index(in_args, GIArgument, metadata->in_arg_index));
-        RVAL2GI_IN_ARGUMENT(argument,
-                            &(metadata->arg_info),
-                            rb_argument,
-                            self);
+        rb_gi_in_argument_from_ruby(callable_info,
+                                    argument,
+                                    &(metadata->arg_info),
+                                    metadata->rb_arg_index,
+                                    rb_argument,
+                                    self);
     }
 }
 
@@ -1215,8 +1316,12 @@ arguments_from_ruby(GICallableInfo *info, VALUE self, VALUE rb_arguments,
 
         metadata = g_ptr_array_index(args_metadata, i);
         if (metadata->in_arg_index != -1) {
-            in_argument_from_ruby(metadata, rb_arguments,
-                                  in_args, args_metadata, self);
+            in_argument_from_ruby(info,
+                                  metadata,
+                                  rb_arguments,
+                                  in_args,
+                                  args_metadata,
+                                  self);
         } else {
             out_argument_from_ruby(metadata, out_args);
         }
@@ -1412,7 +1517,7 @@ rb_gi_function_info_invoke_raw(GIFunctionInfo *info,
             object_class = g_type_class_ref(CLASS2GTYPE(rb_receiver));
             receiver.v_pointer = object_class;
         } else {
-            receiver.v_pointer = DATA_PTR(rb_receiver);
+            receiver.v_pointer = rb_gi_struct_get_raw(rb_receiver, G_TYPE_NONE);
         }
     }
     rb_arguments = rbg_to_array(rb_arguments);
