@@ -1,4 +1,4 @@
-# Copyright (C) 2012-2015  Ruby-GNOME2 Project Team
+# Copyright (C) 2012-2019  Ruby-GNOME2 Project Team
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -37,7 +37,10 @@ module GObjectIntrospection
       repository.require(namespace, @version)
       pre_load(repository, namespace)
       repository.each(namespace) do |info|
-        load_info(info)
+        load_info(info) if info.is_a?(InterfaceInfo)
+      end
+      repository.each(namespace) do |info|
+        load_info(info) unless info.is_a?(InterfaceInfo)
       end
       post_load(repository, namespace)
     end
@@ -78,45 +81,22 @@ module GObjectIntrospection
     end
 
     def define_module_function(target_module, name, function_info)
-      unlock_gvl = should_unlock_gvl?(function_info, target_module)
-      prepare = lambda do |arguments, &block|
-        arguments, block = build_arguments(function_info, arguments, &block)
-        method_name = "#{target_module}\#.#{name}"
-        validate_arguments(function_info, method_name, arguments)
-        [arguments, block]
-      end
+      function_info.unlock_gvl = should_unlock_gvl?(function_info, target_module)
+      full_method_name = "#{target_module}\#.#{name}"
+      invoker = Invoker.new(function_info, name, full_method_name)
       target_module.module_eval do
         define_method(name) do |*arguments, &block|
-          arguments, block = prepare.call(arguments, &block)
-          function_info.invoke({
-                                 :arguments => arguments,
-                                 :unlock_gvl => unlock_gvl,
-                               },
-                               &block)
+          invoker.invoke(nil, arguments, block)
         end
         module_function(name)
       end
     end
 
     def define_singleton_method(klass, name, info)
-      unlock_gvl = should_unlock_gvl?(info, klass)
-      prepare = lambda do |arguments, &block|
-        arguments, block = build_arguments(info, arguments, &block)
-        validate_arguments(info, "#{klass}.#{name}", arguments)
-        [arguments, block]
-      end
-      singleton_class = (class << klass; self; end)
-      singleton_class.__send__(:define_method, name) do |*arguments, &block|
-        arguments, block = prepare.call(arguments, &block)
-        if block.nil? and info.require_callback?
-          to_enum(name, *arguments)
-        else
-          info.invoke({
-                        :arguments => arguments,
-                        :unlock_gvl => unlock_gvl,
-                      },
-                      &block)
-        end
+      info.unlock_gvl = should_unlock_gvl?(info, klass)
+      invoker = Invoker.new(info, name, "#{klass}.#{name}")
+      klass.singleton_class.__send__(:define_method, name) do |*arguments, &block|
+        invoker.invoke(nil, arguments, block)
       end
     end
 
@@ -126,9 +106,11 @@ module GObjectIntrospection
         klass = self.class.define_struct(info.size, name, @base_module,
                                          :parent => options[:parent])
       else
+        size = info.size
+        size = nil if size.zero?
         klass = self.class.define_class(info.gtype, name, @base_module,
                                         :parent => options[:parent],
-                                        :size   => info.size)
+                                        :size   => size)
       end
       load_fields(info, klass)
       load_methods(info, klass)
@@ -141,7 +123,7 @@ module GObjectIntrospection
         method_infos = info.methods.find_all do |method_info|
           method_info.is_a?(MethodInfo)
         end
-        unless methods.empty?
+        unless method_infos.empty?
           base_class = @base_module.const_get(base_class_name)
           load_method_infos(method_infos, base_class.singleton_class)
         end
@@ -153,6 +135,7 @@ module GObjectIntrospection
     end
 
     def define_boxed(info)
+      return if info.gtype == GLib::Type::NONE
       klass = self.class.define_class(info.gtype, info.name, @base_module)
       _ = klass # TODO: Remove me. It is just for suppressing a warning.
       # TODO
@@ -169,6 +152,8 @@ module GObjectIntrospection
     end
 
     def define_enum(info)
+      # TODO: Can we do the same things for flags on NONE GType?
+      return if info.gtype == GLib::Type::NONE
       self.class.define_class(info.gtype,
                               rubyish_class_name(info),
                               @base_module)
@@ -223,6 +208,7 @@ module GObjectIntrospection
     end
 
     def load_object_info(info)
+      return if info.gtype == GLib::Type::NONE
       klass = self.class.define_class(info.gtype,
                                       rubyish_class_name(info),
                                       @base_module)
@@ -241,29 +227,48 @@ module GObjectIntrospection
       field_info.name
     end
 
-    def load_field(info, i, field_info, klass)
+    def load_field(info, i, field_info, klass, options={})
       name = field_name(field_info, klass)
       flags = field_info.flags
 
-      if flags.readable?
-        if field_info.type.tag == TypeTag::BOOLEAN
-          reader_method_name = "#{name}?"
-        else
-          reader_method_name = name
-        end
-        remove_existing_method(klass, reader_method_name)
-        klass.__send__(:define_method, reader_method_name) do ||
-          info.get_field_value(self, i)
-        end
+      readable = options[:readable]
+      readable = flags.readable? if readable.nil?
+      if readable
+        reader_method_name = rubyish_field_reader_name(field_info, name)
+        load_field_reader(info, i, field_info, klass, name, reader_method_name)
       end
 
-      if flags.writable?
-        klass.__send__(:define_method, "#{name}=") do |value|
-          info.set_field_value(self, i, value)
+      writable = options[:writable]
+      writable = flags.writable? if writable.nil?
+      if writable
+        load_field_writer(info, i, field_info, klass, name, "#{name}=")
+      end
+    end
+
+    def load_field_reader(info, i, field_info, klass, name, method_name)
+      remove_existing_method(klass, method_name)
+
+      need_number_to_bool_convert = false
+      if name.start_with?("is_") and
+          field_info.type.tag == TypeTag::UINT32
+        need_number_to_bool_convert = true
+      end
+      klass.__send__(:define_method, method_name) do ||
+        value = info.get_field_value(self, i)
+        if need_number_to_bool_convert
+          value != 0
+        else
+          value
         end
       end
     end
 
+    def load_field_writer(info, i, field_info, klass, name, method_name)
+      klass.__send__(:define_method, method_name) do |value|
+        info.set_field_value(self, i, value)
+      end
+    end
+
     def load_methods(info, klass)
       grouped_methods = info.methods.group_by do |method_info|
         method_info.class
@@ -281,30 +286,24 @@ module GObjectIntrospection
           raise "TODO: #{method_info_class}"
         end
       end
+      if klass.method_defined?(:each) and
+          klass.instance_method(:each).owner == klass
+        klass.include(Enumerable)
+      end
     end
 
     def load_constructor_infos(infos, klass)
       return if infos.empty?
 
-      prepare = lambda do |info, method_name, arguments, &block|
-        arguments, block = build_arguments(info, arguments, &block)
-        validate_arguments(info, "#{klass}\##{method_name}", arguments)
-        [arguments, block]
-      end
       call_initialize_post = lambda do |object|
         initialize_post(object)
       end
       infos.each do |info|
         name = "initialize_#{info.name}"
-        unlock_gvl = should_unlock_gvl?(info, klass)
+        info.unlock_gvl = should_unlock_gvl?(info, klass)
+        invoker = Invoker.new(info, name, "#{klass}\##{name}")
         klass.__send__(:define_method, name) do |*arguments, &block|
-          arguments, block = prepare.call(info, name, arguments, &block)
-          info.invoke({
-                        :receiver  => self,
-                        :arguments => arguments,
-                        :unlock_gvl => unlock_gvl,
-                      },
-                      &block)
+          invoker.invoke(self, arguments, block)
           call_initialize_post.call(self)
         end
         klass.__send__(:private, name)
@@ -322,30 +321,6 @@ module GObjectIntrospection
     def initialize_post(object)
     end
 
-    def build_arguments(info, arguments, &block)
-      last_in_arg = info.in_args.last
-      if block and last_in_arg and last_in_arg.gclosure?
-        [arguments + [block], nil]
-      else
-        [arguments, block]
-      end
-    end
-
-    def validate_arguments(info, method_name, arguments)
-      n_in_args = info.n_in_args
-      n_required_in_args = info.n_required_in_args
-      return if (n_required_in_args..n_in_args).cover?(arguments.size)
-
-      detail = "#{arguments.size} for "
-      if n_in_args == n_required_in_args
-        detail << "#{n_in_args}"
-      else
-        detail << "#{n_required_in_args}..#{n_in_args}"
-      end
-      message = "#{method_name}: wrong number of arguments (#{detail})"
-      raise ArgumentError, message
-    end
-
     def find_suitable_callable_info(infos, arguments)
       min_n_args = nil
       max_n_args = nil
@@ -380,67 +355,9 @@ module GObjectIntrospection
     end
 
     def match_argument?(arg_info, argument)
-      type = arg_info.type
-      case type.tag
-      when TypeTag::BOOLEAN
-        argument == true or argument == false
-      when TypeTag::INT8,
-           TypeTag::UINT8,
-           TypeTag::INT16,
-           TypeTag::UINT16,
-           TypeTag::INT32,
-           TypeTag::UINT32,
-           TypeTag::INT64,
-           TypeTag::UINT64,
-           TypeTag::FLOAT,
-           TypeTag::DOUBLE
-        argument.is_a?(Numeric)
-      when TypeTag::GTYPE
-        argument.is_a?(GLib::Type)
-      when TypeTag::UTF8
-        argument.is_a?(String)
-      when TypeTag::FILENAME
-        argument.is_a?(String)
-      when TypeTag::ARRAY
-        argument.is_a?(Array)
-      when TypeTag::INTERFACE
-        interface = type.interface
-        case interface.type
-        when InfoType::STRUCT
-          match_argument_interface_struct?(arg_info, interface, argument)
-        when InfoType::OBJECT,
-             InfoType::INTERFACE,
-             InfoType::FLAGS,
-             InfoType::ENUM
-          argument.is_a?(interface.gtype.to_class)
-        else
-          # TODO
-          false
-        end
-      when TypeTag::GLIST,
-           TypeTag::GSLIST
-        argument.is_a?(Array)
-      else
-        # TODO
-        false
-      end
-    end
+      return true if argument.nil? and arg_info.may_be_null?
 
-    def match_argument_interface_struct?(arg_info, interface, argument)
-      gtype = interface.gtype
-      case gtype.name
-      when "void"
-        # TODO
-        false
-      when "CairoSurface"
-        if Object.const_defined?(:Cairo)
-          argument.is_a?(Cairo::Surface)
-        else
-          false
-        end
-      else
-        argument.is_a?(gtype.to_class)
-      end
+      arg_info.type.match?(argument)
     end
 
     def rubyish_method_name(function_info, options={})
@@ -448,15 +365,67 @@ module GObjectIntrospection
       if options[:prefix]
         name = name.gsub(/\A#{Regexp.escape(options[:prefix])}/, "")
       end
+      if name == "initialize"
+        name += "_raw"
+      end
+
+      n_in_args = function_info.n_in_args
+      if options[:n_in_args_offset]
+        n_in_args += options[:n_in_args_offset]
+      end
+
       return_type = function_info.return_type
       return_type_tag = return_type.tag
+      if return_type_tag == TypeTag::VOID
+        out_arg_tags = function_info.out_args.collect {|arg| arg.type.tag}
+        if out_arg_tags == [TypeTag::ARRAY, TypeTag::INT32]
+          return_type_tag = TypeTag::ARRAY
+        end
+      end
+
       case return_type_tag
       when TypeTag::BOOLEAN
         case name
-        when /\A(?:is|get_is)_/
+        when "equal"
+          if n_in_args == 1
+            "=="
+          else
+            name
+          end
+        when "not_equal"
+          if n_in_args == 1
+            "!="
+          else
+            name
+          end
+        when "less_than"
+          if n_in_args == 1
+            "<"
+          else
+            name
+          end
+        when "less_than_or_equal"
+          if n_in_args == 1
+            "<="
+          else
+            name
+          end
+        when "greater_than"
+          if n_in_args == 1
+            ">"
+          else
+            name
+          end
+        when "greater_than_or_equal"
+          if n_in_args == 1
+            ">="
+          else
+            name
+          end
+        when /\A(?:is|get_is|can_be)_/
           "#{$POSTMATCH}?"
         when /\Aget_/
-          if function_info.n_in_args.zero?
+          if n_in_args.zero?
             if function_info.n_out_args.zero?
               "#{$POSTMATCH}?"
             else
@@ -472,10 +441,10 @@ module GObjectIntrospection
         else
           name
         end
-      when TypeTag::GLIST, TypeTag::GSLIST
+      when TypeTag::GLIST, TypeTag::GSLIST, TypeTag::ARRAY
         case name
         when /\A(?:list|get)_/
-          if function_info.n_in_args.zero?
+          if n_in_args.zero?
             $POSTMATCH
           else
             name
@@ -486,13 +455,15 @@ module GObjectIntrospection
       else
         case name
         when /\Aget_/
-          if function_info.n_in_args.zero?
+          if n_in_args.zero?
             $POSTMATCH
           else
             name
           end
         when "to_string"
           "to_s"
+        when "to_integer"
+          "to_i"
         when "foreach"
           "each"
         else
@@ -501,6 +472,21 @@ module GObjectIntrospection
       end
     end
 
+    def rubyish_field_reader_name(field_info, name)
+      case field_info.type.tag
+      when TypeTag::BOOLEAN
+        name.gsub(/\Ais_/, "") + "?"
+      when TypeTag::UINT32
+        if /\Ais_/ =~ name
+          "#{$POSTMATCH}?"
+        else
+          name
+        end
+      else
+        name
+      end
+    end
+
     def rubyish_class_name(info)
       name = info.name
       case info
@@ -530,6 +516,7 @@ module GObjectIntrospection
     def load_method_info(info, klass, method_name)
       define_method(info, klass, method_name)
       define_equal_style_setter(info, klass, method_name)
+      define_inspect(info, klass, method_name)
     end
 
     def remove_existing_method(klass, method_name)
@@ -539,46 +526,30 @@ module GObjectIntrospection
     end
 
     def define_method(info, klass, method_name)
-      unlock_gvl = should_unlock_gvl?(info, klass)
+      info.unlock_gvl = should_unlock_gvl?(info, klass)
       remove_existing_method(klass, method_name)
-      function_info_p = (info.class == FunctionInfo)
-      no_return_value_p =
-        (info.return_type.tag == TypeTag::VOID and info.n_out_args.zero?)
-      prepare = lambda do |arguments, &block|
-        arguments, block = build_arguments(info, arguments, &block)
-        validate_arguments(info, "#{klass}\##{method_name}", arguments)
-        [arguments, block]
-      end
-      require_callback_p = info.require_callback?
+      invoker = Invoker.new(info, method_name, "#{klass}\##{method_name}")
       klass.__send__(:define_method, method_name) do |*arguments, &block|
-        arguments = [self] + arguments if function_info_p
-        arguments, block = prepare.call(arguments, &block)
-        if block.nil? and require_callback_p
-          to_enum(method_name, *arguments)
-        else
-          options = {
-            :arguments => arguments,
-            :unlock_gvl => unlock_gvl,
-          }
-          options[:receiver] = self unless function_info_p
-          return_value = info.invoke(options, &block)
-          if no_return_value_p
-            self
-          else
-            return_value
-          end
-        end
+        invoker.invoke(self, arguments, block)
       end
     end
 
     def define_equal_style_setter(info, klass, method_name)
-      if /\Aset_/ =~ method_name and info.n_args == 1
+      if /\Aset_/ =~ method_name and info.n_in_args == 1
         setter_method_name = "#{$POSTMATCH}="
         remove_existing_method(klass, setter_method_name)
         klass.__send__(:alias_method, setter_method_name, method_name)
       end
     end
 
+    def define_inspect(info, klass, method_name)
+      if method_name == "to_s" and info.n_args.zero?
+        klass.__send__(:define_method, "inspect") do ||
+          super().gsub(/\>\z/) {" #{to_s}>"}
+        end
+      end
+    end
+
     def load_function_infos(infos, klass)
       infos.each do |info|
         name = rubyish_method_name(info)
@@ -589,6 +560,7 @@ module GObjectIntrospection
     end
 
     def load_interface_info(info)
+      return if info.gtype == GLib::Type::NONE
       interface_module =
         self.class.define_interface(info.gtype,
                                     rubyish_class_name(info),
@@ -605,9 +577,132 @@ module GObjectIntrospection
     end
 
     def load_union_info(info)
+      return if info.gtype == GLib::Type::NONE
       klass = self.class.define_class(info.gtype, info.name, @base_module)
       load_fields(info, klass)
       load_methods(info, klass)
     end
+
+    class Invoker
+      def initialize(info, method_name, full_method_name)
+        @info = info
+        @method_name = method_name
+        @full_method_name = full_method_name
+        @prepared = false
+      end
+
+      def invoke(receiver, arguments, block)
+        ensure_prepared
+
+        if receiver and @function_info_p
+          arguments.unshift(receiver)
+        end
+
+        arguments, block = build(receiver, arguments, block)
+        unless valid?(arguments)
+          if @on_invalid == :fallback
+            return @value_on_invalid
+          else
+            raise ArgumentError, invalid_error_message(arguments)
+          end
+        end
+
+        if block.nil? and @require_callback_p
+          receiver.to_enum(@method_name, *arguments)
+        else
+          if @function_info_p
+            return_value = @info.invoke(arguments, &block)
+          else
+            return_value = @info.invoke(receiver, arguments, &block)
+          end
+          if @have_return_value_p
+            return_value
+          else
+            receiver
+          end
+        end
+      end
+
+      private
+      def ensure_prepared
+        return if @prepared
+
+        @in_args = @info.in_args
+        @n_in_args = @in_args.size
+        @n_required_in_args = @info.n_required_in_args
+        @last_in_arg = @in_args.last
+        if @last_in_arg
+          @last_in_arg_is_gclosure = @last_in_arg.gclosure?
+        else
+          @last_in_arg_is_gclosure = false
+        end
+        @valid_n_args_range = (@n_required_in_args..@n_in_args)
+
+        @in_arg_types = []
+        @in_arg_nil_indexes = []
+        @in_args.each_with_index do |arg, i|
+          @in_arg_types << arg.type
+          @in_arg_nil_indexes << i if arg.may_be_null?
+        end
+
+        @function_info_p = (@info.class == FunctionInfo)
+        @have_return_value_p = @info.have_return_value?
+        @require_callback_p = @info.require_callback?
+
+        prepare_on_invalid
+
+        @prepared = true
+      end
+
+      def prepare_on_invalid
+        case @method_name
+        when "=="
+          @value_on_invalid = false
+          @on_invalid = :fallback
+        when "!="
+          @value_on_invalid = true
+          @on_invalid = :fallback
+        else
+          @on_invalid = :raise
+        end
+      end
+
+      def build(receiver, arguments, block)
+        if block and @last_in_arg_is_gclosure
+          arguments << block
+          block = nil
+        end
+
+        n_missing_arguments = @n_in_args - arguments.size
+        if 0 < n_missing_arguments
+          @in_arg_nil_indexes.each_with_index do |nil_index, i|
+            next if i <= n_missing_arguments
+            arguments.insert(nil_index, nil)
+          end
+        end
+
+        return arguments, block
+      end
+
+      def valid?(arguments)
+        return false unless @valid_n_args_range.cover?(arguments.size)
+        if @on_invalid == :fallback
+          arguments.zip(@in_arg_types) do |argument, type|
+            return false unless type.match?(argument)
+          end
+        end
+        true
+      end
+
+      def invalid_error_message(arguments)
+        detail = "#{arguments.size} for "
+        if @n_in_args == @n_required_in_args
+          detail << "#{@n_in_args}"
+        else
+          detail << "#{@n_required_in_args}..#{@n_in_args}"
+        end
+        "#{@full_method_name}: wrong number of arguments (#{detail})"
+      end
+    end
   end
 end
